@@ -1,3 +1,4 @@
+import asyncio
 import json
 from datetime import UTC, datetime
 from typing import Any, AsyncGenerator, Callable, Generic, Self, Type, TypeVar, get_origin, overload
@@ -7,7 +8,17 @@ from pydantic import AliasChoices, AliasGenerator, AwareDatetime, BaseModel, Con
 from pydantic.alias_generators import to_camel
 
 from .endpoints.base_fields import IdField
-from .exceptions import SWAPIError, SWAPIException, SWFilterException, SWNoClientProvided
+from .exceptions import (
+    SWAPIError,
+    SWAPIErrorList,
+    SWAPIException,
+    SWAPIGatewayTimeout,
+    SWAPIInternalServerError,
+    SWAPIServiceUnavailable,
+    SWAPITooManyRequests,
+    SWFilterException,
+    SWNoClientProvided,
+)
 from .logging import logger
 
 EndpointClass = TypeVar("EndpointClass", bound="EndpointBase[Any]")
@@ -22,6 +33,7 @@ class ConfigBase:
 class ClientBase:
     api_url: str
     raw: bool
+    language_id: IdField | None = None
 
     def __init__(self, config: ConfigBase, raw: bool = False):
         self.api_url = config.url
@@ -35,21 +47,40 @@ class ClientBase:
         await self._get_client().aclose()
 
     async def log_request(self, request: httpx.Request) -> None:
-        logger.debug(f"Request: {request.method} {request.url} - {request.content!r} <headers: {request.headers}>")
+        if not hasattr(request, "_content"):
+            await request.aread()
+        logger.debug("Request: %s %s - %r <headers: %s>", request.method, request.url, request.content, request.headers)
 
     async def log_response(self, response: httpx.Response) -> None:
         await response.aread()
         logger.debug(
-            f"Response: {response.status_code} - {response.status_code} - {response.content!r} "
-            f"<headers: {response.headers}>"
+            "Response: %s - %s - %r <headers: %s>",
+            response.status_code,
+            response.reason_phrase,
+            response.content,
+            response.headers,
         )
 
     def _get_client(self) -> httpx.AsyncClient:
         raise NotImplementedError()
 
-    @staticmethod
-    def _get_headers() -> dict[str, str]:
-        return {"Content-Type": "application/json", "Accept": "application/json"}
+    def _get_headers(self) -> dict[str, str]:
+        headers = {"Content-Type": "application/json", "Accept": "application/json"}
+
+        if self.language_id is not None:
+            headers["sw-language-id"] = str(self.language_id)
+
+        return headers
+
+    @property
+    def timeout(self) -> httpx.Timeout:
+        client = self._get_client()
+        return client.timeout
+
+    @timeout.setter
+    def timeout(self, timeout: httpx._types.TimeoutTypes) -> None:
+        client = self._get_client()
+        client.timeout = timeout  # type: ignore 
 
     async def _make_request(self, method: str, relative_url: str, **kwargs: Any) -> httpx.Response:
         if relative_url.startswith("http://") or relative_url.startswith("https://"):
@@ -61,44 +92,77 @@ class ClientBase:
         headers = self._get_headers()
         headers.update(kwargs.pop("headers", {}))
 
-        response = await client.request(method, url, headers=headers, **kwargs)
+        retries = int(kwargs.pop("retries", 0))
+        retry_errors = tuple(
+            kwargs.pop("retry_errors", [SWAPIInternalServerError, SWAPIServiceUnavailable, SWAPIGatewayTimeout])
+        )
+        no_retry_errors = tuple(kwargs.pop("no_retry_errors", [SWAPITooManyRequests]))
 
-        if response.status_code >= 400:
-            try:
-                raise SWAPIError(response.json().get("errors"))
-            except json.JSONDecodeError:
-                raise SWAPIError(f"Status: {response.status_code}: {response.text}")
+        retry_count = 0
+        while True:
+            response = await client.request(method, url, headers=headers, **kwargs)
+            if response.status_code >= 400:
+                try:
+                    error: SWAPIError | SWAPIErrorList = SWAPIError.from_errors(response.json().get("errors"))
+                except json.JSONDecodeError:
+                    error: SWAPIError | SWAPIErrorList = SWAPIError.from_response(response)  # type: ignore
 
-        return response
+                if isinstance(error, SWAPIErrorList) and len(error.errors) == 1:
+                    error = error.errors[0]
+
+                if isinstance(error, SWAPIErrorList):
+                    if any([isinstance(err, no_retry_errors) for err in error.errors]):
+                        raise error
+
+                    if not any([isinstance(err, retry_errors) for err in error.errors]):
+                        raise error
+
+                elif isinstance(error, no_retry_errors) or not isinstance(error, retry_errors):
+                    raise error
+
+                if retry_count == retries:
+                    raise error
+
+                logger.debug(f"Try failed, retrying in {2 ** retry_count} seconds.")
+                await asyncio.sleep(2**retry_count)
+                retry_count += 1
+            else:
+                return response
 
     async def get(self, relative_url: str, **kwargs: Any) -> httpx.Response:
         return await self._make_request(method="GET", relative_url=relative_url, **kwargs)
 
     async def post(self, relative_url: str, **kwargs: Any) -> httpx.Response:
         # we need to set a response type, otherwise we don't get one
-        relative_url += "?_response=basic"
+        relative_url += "?_response=basic" if "?" not in relative_url else "&_response=basic"
         return await self._make_request(method="POST", relative_url=relative_url, **kwargs)
 
     async def patch(self, relative_url: str, **kwargs: Any) -> httpx.Response:
         # we need to set a reponse type, otherwise we don't get one
-        relative_url += "?_response=basic"
+        relative_url += "?_response=basic" if "?" not in relative_url else "&_response=basic"
         return await self._make_request(method="PATCH", relative_url=relative_url, **kwargs)
 
     async def delete(self, relative_url: str, **kwargs: Any) -> httpx.Response:
         return await self._make_request(method="DELETE", relative_url=relative_url, **kwargs)
+    
+    async def upload(self, relative_url: str, **kwargs: Any) -> httpx.Response:
+        return await self._make_request(method="POST", relative_url=relative_url, headers={"Content-Type": "application/octet-stream"} ,**kwargs)
 
     async def close(self) -> None:
         await self._get_client().aclose()
 
     async def bulk_upsert(
-        self, name: str, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: dict[str, Any]
+        self, name: str, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: Any
     ) -> dict[str, Any] | None:
         raise SWAPIException("bulk_upsert is only supported in the admin API")
 
     async def bulk_delete(
-        self, name: str, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: dict[str, Any]
+        self, name: str, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: Any
     ) -> dict[str, Any]:
         raise SWAPIException("bulk_delete is only supported in the admin API")
+    
+    def set_language(self, language_id: IdField | None) -> None:
+        self.language_id = language_id
 
 
 class ApiModelBase(BaseModel, Generic[EndpointClass]):
@@ -179,7 +243,7 @@ class ApiModelBase(BaseModel, Generic[EndpointClass]):
         endpoint: EndpointClass = getattr(self._get_client(), self._identifier).__class__(self._get_client())  # type: ignore
         return endpoint
 
-    async def save(self, force_insert: bool = False) -> Self | dict:
+    async def save(self, force_insert: bool = False) -> Self | dict | None:
         endpoint = self._get_endpoint()
 
         if force_insert or self.id is None:
@@ -210,6 +274,7 @@ class EndpointBase(Generic[ModelClass]):
         self.raw = client.raw
         self._filter: list[dict[str, Any]] = []
         self._limit: int | None = None
+        self._page: int | None = None
         self._sort: list[dict[str, Any]] = []
         self._associations: dict[str, dict[str, Any]] = {}
         self._includes: dict[str, list[str]] = {}
@@ -229,6 +294,9 @@ class EndpointBase(Generic[ModelClass]):
         if self._limit is not None:
             data["limit"] = self._limit
 
+        if self._page is not None:
+            data["page"] = self._page
+
         if self._associations:
             data["associations"] = self._associations
 
@@ -240,22 +308,34 @@ class EndpointBase(Generic[ModelClass]):
     def _reset_endpoint(self) -> None:
         self._filter = []
         self._limit = None
+        self._page = None
         self._sort = []
         self._associations = {}
         self._includes = {}
 
     def _serialize_field_name(self, name: str) -> str:
-        return self.model_class.model_fields[name].serialization_alias or name
+        from .endpoints.relations import ForeignRelation, ManyRelation
+
+        # if a model contains ForwardRefs it may be not complete and missing some aliases
+        if not self.model_class.__pydantic_complete__:
+            self.model_class.model_rebuild()
+
+        field = self.model_class.model_fields[name]
+
+        if get_origin(field.annotation) in [ForeignRelation, ManyRelation]:
+            return to_camel(name)
+        else:
+            return self.model_class.model_fields[name].serialization_alias or name
 
     @overload
     def _parse_response(self, data: list[dict[str, Any]]) -> list[ModelClass]:
         # typing overload
-        pass
+        ...
 
     @overload
     def _parse_response(self, data: dict[str, Any]) -> ModelClass:
         # typing overload
-        pass
+        ...
 
     def _parse_response(self, data: list[dict[str, Any]] | dict[str, Any]) -> list[ModelClass] | ModelClass:
         single = False
@@ -267,12 +347,7 @@ class EndpointBase(Generic[ModelClass]):
         result_list: list[ModelClass] = []
 
         for entry in data:
-            api_type = entry.get("type", None)
-
-            if api_type is None:
-                model_class = self.model_class
-            else:
-                model_class = getattr(self.client, api_type).model_class
+            model_class = self.model_class
 
             if "attributes" in entry:
                 obj = model_class(client=self.client, id=entry["id"], **entry["attributes"])
@@ -330,13 +405,17 @@ class EndpointBase(Generic[ModelClass]):
 
         return self._parse_response(result_data)
 
-    async def update(self, pk: str, obj: ModelClass | dict[str, Any]) -> ModelClass | dict[str, Any]:
+    async def update(self, pk: str, obj: ModelClass | dict[str, Any]) -> ModelClass | dict[str, Any] | None:
         if isinstance(obj, ApiModelBase):
             data = obj.model_dump_json(by_alias=True)
         else:
             data = json.dumps(obj)
 
         result = await self.client.patch(f"{self.path}/{pk}", data=data)
+        # 204 - "no data" handling
+        if result.status_code == 204:
+            return None
+
         result_data: dict[str, Any] = self._prase_data_single(result.json())
 
         if self.raw:
@@ -356,13 +435,17 @@ class EndpointBase(Generic[ModelClass]):
 
         return result[0]
 
-    async def create(self, obj: ModelClass | dict[str, Any]) -> ModelClass | dict[str, Any]:
+    async def create(self, obj: ModelClass | dict[str, Any]) -> ModelClass | dict[str, Any] | None:
         if isinstance(obj, ApiModelBase):
             data = obj.model_dump_json(by_alias=True)
         else:
             data = json.dumps(obj)
 
         result = await self.client.post(f"{self.path}", data=data)
+        # 204 - "no data" handling
+        if result.status_code == 204:
+            return None
+
         result_data: dict[str, Any] = self._prase_data_single(result.json())
 
         if self.raw:
@@ -396,7 +479,7 @@ class EndpointBase(Generic[ModelClass]):
         self._includes.update({self._serialize_field_name(field): data for field, data in kwargs.items()})
         return self
 
-    def filter(self, **kwargs: str) -> Self:
+    def filter(self, **kwargs: Any) -> Self:
         for key, value in kwargs.items():
             filter_term = ""
             filter_type = "equals"
@@ -458,18 +541,20 @@ class EndpointBase(Generic[ModelClass]):
         return self
 
     async def bulk_upsert(
-        self, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: dict[str, Any]
+        self, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: Any
     ) -> dict[str, Any] | None:
         return await self.client.bulk_upsert(self.name, objs, **request_kwargs)
 
-    async def bulk_delete(
-        self, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: dict[str, Any]
-    ) -> dict[str, Any]:
+    async def bulk_delete(self, objs: list[ModelClass] | list[dict[str, Any]], **request_kwargs: Any) -> dict[str, Any]:
         return await self.client.bulk_delete(self.name, objs, **request_kwargs)
 
     def limit(self, count: int | None) -> "Self":
         self._limit = count
         return self
+    
+    def page(self, num: int | None) -> "Self":
+        self._page = num
+        return self
 
     def order_by(self, fields: str | tuple[str]) -> "Self":
         if isinstance(fields, str):
